local addon = CreateFrame("Frame")
_G.Broker_Imba = addon

-----------------------------------------------------------------------
-- Raw data
--

local abilityMap = {
	["+5% Critical Chance"] = 1,
	["+10% Attack Power"] = 2,
	["+6% Spell Power"] = 4,
	["+10% Spell Power"] = 8,
	["+10% Physical Haste"] = 16,
	["+5% Spell Haste"] = 32,
	["+5% Primary Stats"] = 64,
	["+3% All Damage"] = 128,
	["Increased Str+Agi"] = 256,
	["Increased Stamina"] = 512,
	["Bloodlust/Heroism"] = 1024,
	["Increased Mana Pool"] = 2048,
	["Increased Mana Regen"] = 4096,
	["Replenishment"] = 8192,

	["+5% Spell Crit. Vulerability"] = 16384,
	["+4% Physical Dmg Taken"] = 32768,
	["+8% Spell Dmg Taken"] = 65536,
	["+30% Bleed Dmg Taken"] = 131072,
	["Decreased Attack Speed"] = 262144,
	["Decreased Cast Speed"] = 524288,
	["Decreased Physical Dmg"] = 1048576,
	["-12% Armor"] = 2097152,
	["-25% Healing Received"] = 4194304,
}

local totalBits = {
	HUNTER = { 4194304, 4194432, 4194306, 4194320 },
	WARRIOR = { 262912, 4621056, 4457217, 3408384 },
	SHAMAN = { 5428, 5437, 5430, 5428 },
	MAGE = { 3076, 527492, 19468, 11268 },
	PRIEST = { 512, 512, 512, 4203040 },
	WARLOCK = { 589824, 589824, 589832, 614400 },
	ROGUE = { 6815744, 6881280, 6848512, 6946817 },
	DRUID = { 64, 65632, 3539009, 8256 },
	PALADIN = { 4162, 4162, 1314882, 12482 },
	DEATHKNIGHT = { 262402, 1310978, 295186, 327938 },
	Devilsaur = 4194305,
	Fox = 262144,
	Worm = 32768,
	Cat = 256,
	Ravager = 32768,
	Wolf = 1,
	Serpent = 2097152,
	["Shale Spider"] = 64,
	Bear = 1048576,
	Tallstrider = 262144,
	Sporebat = 524288,
	["Core Hound"] = 525312,
	Dragonhawk = 65536,
	Boar = 131072,
	Silithid = 512,
	["Wind Serpent"] = 65536,
	Rhino = 131072,
	Raptor = 2097152,
	Felhunter = 6144,
	["Spirit Beast"] = 256,
	Hyena = 131072,
}
local L = GetLocale()
if L == "deDE" then
	totalBits.GermanWordForFox = totalBits.Fox
	-- etc...
end

-----------------------------------------------------------------------
-- NotifyInspect damage control
--

local lastUnitQueried = nil

hooksecurefunc("NotifyInspect", function(unit)
	-- If we queried this unit just return
	if unit == lastUnitQueried then return end
	-- Someone else queried this unit; time to do some damage control.
	--addon:UnregisterEvent("INSPECT_READY")
	-- XXX Need to figure out if this is needed now with the talent
	-- API changes. Maybe Blizzard made a smart move and included
	-- unit name or whatever as an argument to INSPECT_READY?
	-- I have no idea. But if they did, then yay.
end)
local function queryUnit(unit)
	lastUnitQueried = unit
	NotifyInspect(unit)
end

-----------------------------------------------------------------------
-- Local variables
--

-- Holds a map of group members and bitwise what they provide
local providers = {}
-- Number of abilities that the current group covers (max 23).
local abilitiesCovered = 0
-- Missing abilites (for fast tooltip access)
local missingAbilities = {}

-- Temporary table used several places; make sure to wipe it!
local tmp = {}

-----------------------------------------------------------------------
-- Init
--

local broker = LibStub("LibDataBroker-1.1")
local imba = broker:NewDataObject("Imba", {
	icon = "Interface\\AddOns\\Broker_CMA\\icon",
	type = "data source",
	text = "Imba",
	OnTooltipShow = function(tt)
		if #missingAbilities == 0 then return end
		tt:AddLine("Missing buffs/debuffs")
		for i, ability in next, missingAbilities do
			tt:AddLine(ability, 1, 1, 1, 0, 1, 0)
		end
		tt:AddLine(" ")
		tt:AddLine("Everything is drycoded but you have a nice tooltip here, lucky you!", 0.2, 1, 0.2, 1)
	end,
})
addon:SetScript("OnEvent", function(self, event, ...)
	self[event](self, ...)
end)
addon:RegisterEvent("VARIABLES_LOADED")
addon:RegisterEvent("UNIT_PET")
addon:RegisterEvent("INSPECT_READY")
function addon:VARIABLES_LOADED() end

-----------------------------------------------------------------------
-- Display
--

local bb = bit.band
local function updateTotals()
	wipe(missingAbilities)
	abilitiesCovered = 0

	for ability, B in pairs(abilityMap) do
		for unit, provides in pairs(providers) do
			if bb(provides, B) == B then
				tmp[ability] = true
			end
		end
	end
	for ability in pairs(abilityMap) do
		if not tmp[ability] then
			missingAbilities[#missingAbilities + 1] = ability
		else
			abilitiesCovered = abilitiesCovered + 1
		end
	end
	
	-- If we are at 23/23 we should probably have a token like "OK" instead of counter
	imba.text = ("%d/23"):format(abilitiesCovered)
	wipe(tmp)
end

-----------------------------------------------------------------------
-- Scanning
--

local needsScanning = {}

local total = 0
local lastQueryFired = 0
addon:SetScript("OnUpdate", function(self, elapsed)
	if #needsScanning == 0 then
		self:Hide()
		return
	end

	total = total + elapsed
	if lastUnitQueried then
		-- We are waiting for a query
		local t = GetTime()
		-- We only wait 5 seconds for a result, then we give up.
		if GetTime() > (lastQueryFired + 5) then
			lastUnitQueried = nil
		end
	end
	if lastUnitQueried then return end
	if total > 1 then
		queryUnit(table.remove(needsScanning))
		total = 0
	end
end)

-----------------------------------------------------------------------
-- Roster
--

local function scanRoster()
	if not UnitInRaid("player") then return end
	for i = 1, MAX_RAID_MEMBERS do
		local name = GetRaidRosterInfo(i)
		tmp[name] = true
	end
	for k, v in pairs(providers) do
		if not UnitExists(v) then providers[v] = nil end
		if not v:find("pet$") and not tmp[v] then providers[v] = nil end
	end
	for k, v in pairs(tmp) do
		if not providers[v] then
			needsScanning[#needsScanning + 1] = v
		end
	end
	if #needsScanning > 0 then
		addon:Show()
	end
	wipe(tmp)
end

addon:RegisterEvent("RAID_ROSTER_UPDATE")
addon:RegisterEvent("PARTY_MEMBERS_CHANGED")
addon:RegisterEvent("PLAYER_LOGIN")
function addon:PARTY_MEMBERS_CHANGED() scanRoster() end
addon.RAID_ROSTER_UPDATE = addon.PARTY_MEMBERS_CHANGED
function addon:PLAYER_LOGIN() if GetNumRaidMembers() > 0 then scanRoster() end end

-----------------------------------------------------------------------
-- Provider calculation
--

function addon:INSPECT_READY(...)
	if select("#", ...) > 0 then print(...) end
	local group = GetActiveTalentGroup(true)
	local tree = 0
	for i = 1, 3 do
		local _,_,_,_, pointsSpent = GetTalentTabInfo(i, true, nil, group)
		if pointsSpent > 30 then
			tree = i
			break
		end
	end
	local _, class = UnitClass(lastUnitQueried)
	if not class or not totalBits[class] or not totalBits[class][tree] then return end
	local bits = totalBits[class][tree]
	if providers[lastUnitQueried] ~= bits then
		providers[lastUnitQueried] = totalBits[class][tree]
		updateTotals()
	end
	lastUnitQueried = nil
end

function addon:UNIT_PET(unit)
	local petId = unit.."pet"
	local family = UnitCreatureFamily(petId)
	if not totalBits[family] then return end
	if providers[petId] ~= totalBits[family] then
		providers[petId] = totalBits[family]
		updateTotals()
	end
end

